import datetime
from typing import Dict, List

from django.contrib import messages
from django.contrib.admin.views.decorators import staff_member_required
from django.shortcuts import get_object_or_404, redirect, render, reverse
from django.views.decorators.http import require_http_methods
from django.views.generic import ListView, TemplateView, UpdateView
from formtools.wizard.views import SessionWizardView

from booking.forms import (BookingCustomerForm, BookingDateForm,
                           BookingSettingsForm, BookingTimeForm)
from booking.models import Booking, BookingSettings
from booking.settings import (BOOKING_BG, BOOKING_DESC, BOOKING_DISABLE_URL,
                              BOOKING_SUCCESS_REDIRECT_URL, BOOKING_TITLE,
                              PAGINATION)
from booking.utils import BookingSettingMixin


# # # # # # #
# Admin Part
# # # # # # #
class BookingHomeView(BookingSettingMixin, TemplateView):
    model = Booking
    template_name = "booking/admin/dashboard.html"

    def get_context_data(self, **kwargs):
        context = super().get_context_data(**kwargs)
        context["last_bookings"] = Booking.objects.filter().order_by(
            "date", "time")[:10]
        context["waiting_bookings"] = Booking.objects.filter(
            approved=False).order_by("-date", "time")[:10]
        return context


class BookingListView(BookingSettingMixin, ListView):
    model = Booking
    template_name = "booking/admin/booking_list.html"
    paginate_by = PAGINATION


class BookingSettingsView(BookingSettingMixin, UpdateView):
    form_class = BookingSettingsForm
    template_name = "booking/admin/booking_settings.html"

    def get_object(self):
        return BookingSettings.objects.filter().first()

    def get_success_url(self):
        return reverse("booking_settings")


@require_http_methods(["GET", "POST"])
@staff_member_required
def bookingUpdateView(request, id, type):
    if request.method == "GET":
        item = get_object_or_404(Booking, id=id)
        if type == "delete":
            item.delete()
            messages.warning(request, "The item successfully deleted!")
        elif type == "approved":
            item.approved = True
            item.save()
            messages.success(request, "The item successfully approved!")
        return redirect(reverse("booking_list"))
    return redirect(reverse("create_booking"))


# # # # # # # #
# Booking Part
# # # # # # # #
BOOKING_STEP_FORMS = (
    ('Date', BookingDateForm),
    ('Time', BookingTimeForm),
    ('User Info', BookingCustomerForm)
)


class BookingCreateWizardView(SessionWizardView):
    template_name = "booking/user/booking_wizard.html"
    form_list = BOOKING_STEP_FORMS

    def get_context_data(self, form, **kwargs):
        context = super().get_context_data(form=form, **kwargs)
        progress_width = "6"
        if self.steps.current == 'Time':
            context["get_available_time"] = get_available_time(
                self.get_cleaned_data_for_step('Date')["date"])
            progress_width = "30"
        if self.steps.current == 'User Info':
            progress_width = "75"

        context.update({
            'booking_settings': BookingSettings.objects.first(),
            "progress_width": progress_width,
            "booking_bg": BOOKING_BG,
            "description": BOOKING_DESC,
            "title": BOOKING_TITLE

        })
        return context

    def render(self, form=None, **kwargs):
        # Check if Booking is Disable
        form = form or self.get_form()
        context = self.get_context_data(form=form, **kwargs)

        if not context["booking_settings"].booking_enable:
            return redirect(BOOKING_DISABLE_URL)

        return self.render_to_response(context)

    def done(self, form_list, **kwargs):
        data = dict((key, value) for form in form_list for key,
                    value in form.cleaned_data.items())
        booking = Booking.objects.create(**data)

        if BOOKING_SUCCESS_REDIRECT_URL:
            return redirect(BOOKING_SUCCESS_REDIRECT_URL)

        return render(self.request, 'booking/user/booking_done.html', {
            "progress_width": "100",
            "booking_id": booking.id,
            "booking_bg": BOOKING_BG,
            "description": BOOKING_DESC,
            "title": BOOKING_TITLE
        })


def add_delta(time: datetime.time, delta: datetime.datetime) -> datetime.time:
    # transform to a full datetime first
    return (datetime.datetime.combine(
        datetime.date.today(), time
    ) + delta).time()


def get_available_time(date: datetime.date) -> List[Dict[datetime.time, bool]]:
    """
    Check for all available time for selected date
    The times should ne betwwen start_time and end_time
    If the time already is taken -> is_taken = True
    """
    booking_settings = BookingSettings.objects.first()
    existing_bookings = Booking.objects.filter(
        date=date).values_list('time')

    next_time = booking_settings.start_time
    time_list = []
    while True:
        is_taken = any([x[0] == next_time for x in existing_bookings])
        time_list.append(
            {"time": ":".join(str(next_time).split(":")[:-1]), "is_taken": is_taken})
        next_time = add_delta(next_time, datetime.timedelta(
            minutes=int(booking_settings.period_of_each_booking)))
        if next_time > booking_settings.end_time:
            break

    return time_list



from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

DEMO_DATA = {
    "user_name": "booking_user",
    "email": "booking@test.com",
    "password": "booking_password"
}


class TestPermissions(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            DEMO_DATA["user_name"], DEMO_DATA["email"], DEMO_DATA["password"], is_staff=True)

    def test_admin_view_permissions(self):
        # admin_dashboard
        response = self.client.get(reverse("admin_dashboard"))
        self.assertEqual(response.status_code, 302)
        # booking_list
        response = self.client.get(reverse("booking_list"))
        self.assertEqual(response.status_code, 302)
        # booking_settings
        response = self.client.get(reverse("booking_settings"))
        self.assertEqual(response.status_code, 302)


        # Loggin as stuff user
        self.client.login(
            username=DEMO_DATA["user_name"], password=DEMO_DATA["password"])

        # admin_dashboard
        response = self.client.get(reverse("admin_dashboard"))
        self.assertEqual(response.status_code, 200)
        # booking_list
        response = self.client.get(reverse("booking_list"))
        self.assertEqual(response.status_code, 200)
        # booking_settings
        response = self.client.get(reverse("booking_settings"))
        self.assertEqual(response.status_code, 200)



import datetime

from django.test import TestCase

from booking.forms import (BookingCustomerForm, BookingDateForm,
                           BookingSettingsForm, BookingTimeForm)

DEMO_BOOKING_DATA = {
        "date": (datetime.datetime.now() + datetime.timedelta(days=1)).strftime("%Y-%m-%d"),
        "time": "12:00",
        "user_name": "test_user",
        "user_email": "user@email.com",
        "user_mobile": ""
    }

class TestBookingCreateForms(TestCase):
    data = DEMO_BOOKING_DATA

    def test_date_form(self):
        # Test the valid form
        form = BookingDateForm(data=self.data)
        validated = form.is_valid()
        self.assertTrue(validated)

        # Test the validation
        self.data["date"] = "08:00"
        form = BookingDateForm(data=self.data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(form.errors["date"][0], "Enter a valid date.")

        # Test for requreid fields error
        del self.data["date"]
        form = BookingDateForm(data=self.data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(len(form.errors), 1)
        self.assertEqual(len(form.errors["date"]), 1)
        self.assertEqual(form.errors["date"][0], "This field is required.")

    def test_time_form(self):
        # Test the valid form
        form = BookingTimeForm(data=self.data)
        validated = form.is_valid()
        self.assertTrue(validated)

        # Test the validation
        self.data["time"] = "asd"
        form = BookingTimeForm(data=self.data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(form.errors["time"][0], "Enter a valid time.")

        # Test for requreid fields error
        del self.data["time"]
        form = BookingTimeForm(data=self.data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(len(form.errors), 1)
        self.assertEqual(len(form.errors["time"]), 1)

    def test_customer_form(self):
        # Test the valid form
        form = BookingCustomerForm(data=self.data)
        validated = form.is_valid()
        self.assertTrue(validated)

        # Test the validation
        self.data["user_email"] = "asd"
        form = BookingCustomerForm(data=self.data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(form.errors["user_email"]
                         [0], "Enter a valid email address.")

        # Test for requreid fields error
        del self.data["user_name"]
        del self.data["user_email"]

        form = BookingCustomerForm(data=self.data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(len(form.errors), 2)
        self.assertEqual(form.errors["user_name"]
                         [0], "This field is required.")
        self.assertEqual(form.errors["user_email"]
                         [0], "This field is required.")


# Test the Booking Setting Form
class TestBookingSettingsForm(TestCase):
    def test_settings_form(self):
        data = {
            "start_time": "09:00",
            "end_time": "17:00",
            "booking_enable": True,
            "confirmation_required": True,
            "available_booking_months": 1,
            "max_booking_per_time": 1,
            "period_of_each_booking": "30",
        }

        # Test the valid form
        form = BookingSettingsForm(data=data)
        validated = form.is_valid()
        self.assertTrue(validated)

        # Test the validation
        data["end_time"] = "08:00"
        form = BookingSettingsForm(data=data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(len(form.non_field_errors()), 1)
        self.assertEqual(
            form.non_field_errors()[
                0], "The end time must be later than start time."
        )

        # Test for requreid fields error
        del data["start_time"]
        del data["end_time"]
        form = BookingSettingsForm(data=data)
        validated = form.is_valid()
        self.assertFalse(validated)
        self.assertEqual(len(form.non_field_errors()), 0)
        self.assertEqual(len(form.errors), 2)
        self.assertEqual(len(form.errors["end_time"]), 1)
        self.assertEqual(form.errors["end_time"][0], "This field is required.")
        self.assertEqual(len(form.errors["start_time"]), 1)
        self.assertEqual(form.errors["start_time"]
                         [0], "This field is required.")



from django.conf import settings
from django.contrib.auth.decorators import login_required
from django.shortcuts import redirect
from django.utils.decorators import method_decorator


class BookingSettingMixin(object):
    """
    View mixin which requires that the authenticated user is a staff member
    (i.e. `is_staff` is True).
    """

    @method_decorator(login_required)
    def dispatch(self, request, *args, **kwargs):
        if not request.user.is_staff:
            return redirect(settings.LOGIN_URL)
        return super().dispatch(request, *args, **kwargs)



from django.conf import settings

PAGINATION = getattr(settings, 'DJ_BOOKING_PAGINATION', 10)

BOOKING_SUCCESS_REDIRECT_URL = getattr(settings, 'BOOKING_SUCCESS_REDIRECT_URL', None)

BOOKING_DISABLE_URL = getattr(settings, 'BOOKING_DISABLE_URL', "/")

BOOKING_BG = getattr(settings, 'BOOKING_BG', "img/booking_bg.jpg")

BOOKING_TITLE = getattr(settings, 'BOOKING_TITLE', "Booking")

BOOKING_DESC = getattr(settings, 'BOOKING_DESC', "Make your booking easy and fast with us.")



from django.urls import path

from .views import (BookingHomeView, BookingCreateWizardView, BookingListView,
                    BookingSettingsView, bookingUpdateView, get_available_time)

urlpatterns = [
    path("", BookingCreateWizardView.as_view(), name="create_booking"),
    path("admin", BookingHomeView.as_view(), name="admin_dashboard"),
    path("admin/list", BookingListView.as_view(), name="booking_list"),
    path("admin/settings", BookingSettingsView.as_view(), name="booking_settings"),
    path("<int:id>/<str:type>", bookingUpdateView, name="booking_update"),
    path("get-available-time", get_available_time, name="get_available_time"),
]



from django import forms

from booking.models import BookingSettings


class ChangeInputsStyle(forms.Form):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        # add common css classes to all widgets
        for field in iter(self.fields):
            # get current classes from Meta
            input_type = self.fields[field].widget.input_type
            classes = self.fields[field].widget.attrs.get("class")
            if classes is not None:
                classes += " form-check-input" if input_type == "checkbox" else " form-control  flatpickr-input"
            else:
                classes = " form-check-input" if input_type == "checkbox" else " form-control flatpickr-input"
            self.fields[field].widget.attrs.update({
                'class': classes
            })


class BookingDateForm(ChangeInputsStyle):
    date = forms.DateField(required=True)


class BookingTimeForm(ChangeInputsStyle):
    time = forms.TimeField(widget=forms.HiddenInput())


class BookingCustomerForm(ChangeInputsStyle):
    user_name = forms.CharField(max_length=250)
    user_email = forms.EmailField()
    user_mobile = forms.CharField(required=False, max_length=10)


class BookingSettingsForm(ChangeInputsStyle, forms.ModelForm):
    start_time = forms.TimeField(widget=forms.TimeInput(format='%H:%M'))
    end_time = forms.TimeField(widget=forms.TimeInput(format='%H:%M'))

    def clean(self):
        if "end_time" in self.cleaned_data and "start_time" in self.cleaned_data:
            if self.cleaned_data["end_time"] <= self.cleaned_data["start_time"]:
                raise forms.ValidationError(
                    "The end time must be later than start time."
                )
        return self.cleaned_data

    class Meta:
        model = BookingSettings
        fields = "__all__"
        exclude = [
            # TODO: Add this fields to admin panel and fix the functions
            "max_booking_per_time",
            "max_booking_per_day",
        ]



import os

from setuptools import find_packages, setup

with open(os.path.join(os.path.dirname(__file__), 'README.rst')) as readme:
    README = readme.read()

# allow setup.py to be run from any path
os.chdir(os.path.normpath(os.path.join(os.path.abspath(__file__), os.pardir)))

requirements = [
    "asgiref>=3",
    "backports.zoneinfo>=0.2",
    "Django>=4.",
    "django-formtools>=2",
    "pycodestyle>=2",
    "sqlparse>=0.4",
    "toml>=0.10"
]

setup(
    name='dj-booking',
    version='1.6.2',
    description='Django Booking is a complete Django booking system as a package.',
    long_description=README,
    long_description_content_type='text/markdown',
    url='https://github.com/foad-heidari/dj-booking',
    author='Foad',
    author_email='foad.haydri.1377@gmail.com',
    license='MIT',
    maintainer='Foad',
    maintainer_email='foad.haydri.1377@gmail.com',
    keywords='django booking appointment appointment-booking booking-system appointment-system doctor-appointment-booking django-appointment django-booking',
    packages=find_packages(),
    include_package_data=True,
    install_requires=requirements,
    classifiers=[
        'Environment :: Web Environment',
        "Framework :: Django",
        "Framework :: Django :: 2.2",
        "Framework :: Django :: 3.2",
        "Framework :: Django :: 4.0",
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Internet :: WWW/HTTP',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



from django.contrib.auth.models import User
from django.test import Client, TestCase
from django.urls import reverse
from django.utils.encoding import force_str

from booking.models import Booking
from booking.tests.test_forms import DEMO_BOOKING_DATA

DEMO_USER_DATA = {
    "user_name": "booking_user",
    "email": "booking@test.com",
    "password": "booking_password"
}


class TestBookingAdminViews(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            DEMO_USER_DATA["user_name"], DEMO_USER_DATA["email"], DEMO_USER_DATA["password"],
            is_staff=True,
            is_superuser=True
        )

        self.booking = Booking.objects.create(**DEMO_BOOKING_DATA)

        self.cl = Client()
        # Loggin as stuff user
        self.cl.login(
            username=DEMO_USER_DATA["user_name"], password=DEMO_USER_DATA["password"])

    def test_booking_settings_view(self):

        # booking settings
        response = self.cl.post(
            reverse("booking_settings"),
            {
                "booking_enable": "on",
                "confirmation_required": "on",
                # "disable_weekend": "on",
                "available_booking_months": "1",
                "start_time": "09:00",
                "end_time": "17:00",
                "period_of_each_booking": "30",
            },
        )
        self.assertEqual(response.status_code, 302)

        response = self.cl.get(reverse("booking_settings"))
        self.assertInHTML(
            '<input type="checkbox" name="booking_enable" class=" form-check-input" id="id_booking_enable" checked="">', force_str(response.content))
        self.assertInHTML(
            '<input type="checkbox" name="disable_weekend" class=" form-check-input" id="id_disable_weekend">', force_str(response.content))

    def test_booking_list_view(self):
        # booking List
        response = self.cl.get(reverse("booking_list"))
        self.assertEqual(response.status_code, 200)
        self.assertInHTML(
            f'<td>{DEMO_BOOKING_DATA["user_name"]}</td>', force_str(response.content))
        self.assertInHTML(
            '<i class="far fa-pause-circle text-danger"></i>', force_str(response.content))

    def test_approve_booking_view(self):
        response = self.cl.get(reverse("booking_update", kwargs={
                               "id": self.booking.id, "type": "approved"}))
        self.assertEqual(response.status_code, 302)
        response = self.cl.get(reverse("booking_list"))
        self.assertEqual(response.status_code, 200)
        self.assertInHTML(
            '<i class="fas fa-check-square text-success"></i>', force_str(response.content))

    def test_delete_booking_view(self):
        response = self.cl.get(reverse("booking_update", kwargs={
                               "id": self.booking.id, "type": "delete"}))
        self.assertEqual(response.status_code, 302)
        response = self.cl.get(reverse("booking_list"))
        self.assertEqual(response.status_code, 200)
        self.assertInHTML(
            '<tbody></tbody>', force_str(response.content))



class TestBookingUserViews(TestCase):
    # TODO: Create the test
    pass



# Generated by Django 3.1.7 on 2021-08-08 14:13

from django.db import migrations, models


def forwards_func(apps, schema_editor):
    BookingSettings = apps.get_model("booking", "BookingSettings")
    db_alias = schema_editor.connection.alias
    BookingSettings.objects.using(db_alias).create(start_time="09:00", end_time="17:00")

def reverse_func(apps, schema_editor):
    # remove the object we created on forwards_func
    BookingSettings = apps.get_model("booking", "BookingSettings")
    db_alias = schema_editor.connection.alias
    BookingSettings.objects.using(db_alias).filter(start_time="09:00", end_time="17:00").delete()

class Migration(migrations.Migration):

    dependencies = [
        ('booking', '0001_initial'),
    ]

    operations = [
        migrations.CreateModel(
            name='BookingSettings',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('booking_enable', models.BooleanField(default=True)),
                ('confirmation_required', models.BooleanField(default=True)),
                ('disable_weekend', models.BooleanField(default=True)),
                ('available_booking_months', models.IntegerField(default=1, help_text='if 2, user can only book booking for next two months.')),
                ('max_booking_per_day', models.IntegerField(blank=True, null=True)),
                ('start_time', models.TimeField()),
                ('end_time', models.TimeField()),
                ('period_of_each_booking', models.CharField(choices=[('5', '5M'), ('10', '10M'), ('15', '15M'), ('20', '20M'), ('25', '25M'), ('30', '30M'), ('35', '35M'), ('40', '40M'), ('45', '45M'), ('60', '1H'), ('75', '1H 15M'), ('90', '1H 30M'), ('105', '1H 45M'), ('120', '2H'), ('150', '2H 30M'), ('180', '3H')], default='30', help_text='How long each booking take.', max_length=3)),
                ('max_booking_per_time', models.IntegerField(default=1, help_text='how much booking can be book for each time.')),
            ],
        ),
        # Create base booking_settings for first time
        migrations.RunPython(forwards_func, reverse_func),
    ]
